
#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

import block2
from block2 import *
from block2.su2 import *
from pyscf import lib, ao2mo, gto, mcscf, tools
from pyscf.mcscf import casci_symm
import numpy as np

class DMRGCI(lib.StreamObject):
    """DMRG FCI solver."""

    def __init__(self, mf):
        self.mol = mf.mol
        self._scf = mf
        self.verbose = self.mol.verbose
        self.stdout = self.mol.stdout
        self.e_tot = None
        self.dmrg_args = {
            "startM": 250, "maxM": 500, "schedule": "default",
            "sweep_tol": 1E-6, "cutoff": 1E-14,
            "memory": lib.param.MAX_MEMORY * 1E6
        }
    
    @staticmethod
    def get_schedule(kwargs):
        start_m = int(kwargs.get("startM", 250))
        max_m = int(kwargs.get("maxM", 0))
        if max_m <= 0:
            raise ValueError("A positive maxM must be set for default schedule, "
                            + "current value : %d" % max_m)
        elif max_m < start_m:
            raise ValueError(
                "maxM %d cannot be smaller than startM %d" % (max_m, start_m))
        sch_type = kwargs.get("schedule", "")
        sweep_tol = float(kwargs.get("sweep_tol", 0))
        if sweep_tol == 0:
            kwargs["sweep_tol"] = "1E-5"
            sweep_tol = 1E-5
        schedule = []
        if sch_type == "default":
            def_m = [50, 100, 250, 500] + [1000 * x for x in range(1, 11)]
            def_iter = [8] * 5 + [4] * 9
            def_noise = [1E-3] * 3 + [1E-4] * 2 + [5E-5] * 9
            def_tol = [1E-4] * 3 + [1E-5] * 2 + [5E-6] * 9
            if start_m == max_m:
                schedule.append([0, start_m, 1E-5, 1E-4])
                schedule.append([8, start_m, 5E-6, 5E-5])
            elif start_m < def_m[0]:
                def_m.insert(0, start_m)
                for x in [def_iter, def_noise, def_tol]:
                    x.insert(0, x[0])
            elif start_m > def_m[-1]:
                while start_m > def_m[-1]:
                    def_m.append(def_m[-1] + 1000)
                    for x in [def_iter, def_noise, def_tol]:
                        x.append(x[-1])
            else:
                for i in range(1, len(def_m)):
                    if start_m < def_m[i]:
                        def_m[i - 1] = start_m
                        break
            isweep = 0
            for i in range(len(def_m)):
                if def_m[i] >= max_m:
                    schedule.append([isweep, max_m, def_tol[i], def_noise[i]])
                    isweep += def_iter[i]
                    break
                elif def_m[i] >= start_m:
                    schedule.append([isweep, def_m[i], def_tol[i], def_noise[i]])
                    isweep += def_iter[i]
            schedule.append([schedule[-1][0] + 8, max_m, sweep_tol / 10, 0.0])
            last_iter = schedule[-1][0]
            if "twodot" not in kwargs and "onedot" not in kwargs \
                and "twodot_to_onedot" not in kwargs:
                kwargs["twodot_to_onedot"] = str(last_iter + 2)
            max_iter = int(kwargs.get("maxiter", 0))
            if max_iter <= schedule[-1][0]:
                kwargs["maxiter"] = str(last_iter + 4)
                max_iter = last_iter + 4
        else:
            raise ValueError("Unrecognized schedule type (%s)" % sch_type)
        
        tmp = list(zip(*schedule))
        nsweeps = np.diff(tmp[0]).tolist()
        maxiter = int(kwargs.get("maxiter", 1)) - int(np.sum(nsweeps))
        assert maxiter > 0
        nsweeps.append(maxiter)

        schedule = [[], [], []]
        for nswp, M, tol, noise in zip(nsweeps, *tmp[1:]):
            schedule[0].extend([M] * nswp)
            schedule[1].extend([tol] * nswp)
            schedule[2].extend([noise] * nswp)
        kwargs["schedule"] = schedule
        return schedule
    
    def kernel(self, h1e, g2e, norb, nelec, ecore=0, ci0=None, **kwargs):

        # Global
        Random.rand_seed(123456)
        scratch = './nodex'
        n_threads = lib.num_threads()
        fcidump_tol = 1E-13
        memory = self.dmrg_args["memory"]
        init_memory(isize=int(memory * 0.05),
            dsize=int(memory * 0.95), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global,
            n_threads, n_threads, 1)
        Global.threading.seq_type = SeqTypes.Tasked
        Global.frame.fp_codec = DoubleFPCodec(1E-16, 1024)
        Global.frame.load_buffering = False
        Global.frame.save_buffering = False
        Global.frame.use_main_stack = False
        Global.frame.minimal_disk_usage = True
        if self.verbose >= 6:
            print(Global.frame)
            print(Global.threading)

        if 'orbsym' in kwargs:
            orb_sym = kwargs['orbsym']
        else:
            orb_sym = [0] * norb

        n_orbs = norb
        self.n_orbs = n_orbs
        mp_orb_sym = [tools.fcidump.ORBSYM_MAP[self.mol.groupname][i] for i in orb_sym]
        h1e = h1e.flatten()
        g2e = ao2mo.restore(8, g2e, n_orbs)
        h1e[np.abs(h1e) < fcidump_tol] = 0
        g2e[np.abs(g2e) < fcidump_tol] = 0
        na, nb = nelec
        fcidump = FCIDUMP()
        fcidump.initialize_su2(n_orbs, na + nb, na - nb, 1, ecore, h1e, g2e)
        fcidump.orb_sym = VectorUInt8(mp_orb_sym)
        error = fcidump.symmetrize(VectorUInt8(orb_sym))
        if self.verbose >= 6:
            print('symm error = ', error)

        if self.dmrg_args["schedule"] == "default":
            DMRGCI.get_schedule(self.dmrg_args)
        
        bond_dims, dav_thrds, noises = self.dmrg_args["schedule"]
        sweep_tol = self.dmrg_args["sweep_tol"]
        
        # Hamiltonian
        vacuum = SU2(0)
        target = SU2(na + nb, abs(na - nb), 0)
        hamil = HamiltonianQC(vacuum, fcidump.n_sites, VectorUInt8(orb_sym), fcidump)
        n_sites = hamil.n_sites

        self.hamil = hamil

        # MPS
        info = MPSInfo(n_sites, vacuum, target, hamil.basis)
        info.set_bond_dimension(bond_dims[0])
        mps = MPS(n_sites, 0, 2)
        mps.initialize(info)
        mps.random_canonicalize()
        mps.tensors[mps.center].normalize()
        mps.save_mutable()
        info.save_mutable()
        forward = mps.center == 0

        # MPO
        mpo = MPOQC(hamil, QCTypes.NC)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True, OpNamesSet((OpNames.R, OpNames.RD)))

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        me.save_partition_info = True
        me.init_environments(self.verbose >= 5)
        dmrg = DMRG(me, VectorUBond(bond_dims), VectorDouble(noises))
        dmrg.davidson_conv_thrds = VectorDouble(dav_thrds)
        dmrg.cutoff = self.dmrg_args["cutoff"]
        dmrg.iprint = max(min(self.verbose - 4, 3), 0)
        dmrg.noise_type = NoiseTypes.ReducedPerturbativeCollected
        if "twodot_to_onedot" not in self.dmrg_args:
            self.e_tot = dmrg.solve(len(bond_dims), forward, sweep_tol)
        else:
            tto = int(self.dmrg_args["twodot_to_onedot"])
            assert len(bond_dims) > tto
            dmrg.solve(tto, forward, 0)
            dmrg.bond_dims = VectorUBond(bond_dims[tto:])
            dmrg.noises = VectorDouble(noises[tto:])
            dmrg.davidson_conv_thrds = VectorDouble(dav_thrds[tto:])
            dmrg.me.dot = 1
            self.e_tot = dmrg.solve(len(bond_dims) - tto, mps.center == 0, sweep_tol)
            mps.dot = 1
            mps.save_data()

        self.mps = mps
        self.converged = True
        return self.e_tot, mps

    def make_rdm1(self, state=None, norb=None, nelec=None):
        '''
        Spin-traced one-particle density matrix in MO basis.
        dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>
        '''

        if state is None:
            state = self.mps
            state.dot = 2
        
        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil, 0)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 1PDM
        pme = MovingEnvironment(pmpo, state, state, "1PDM")
        pme.init_environments(self.verbose >= 5)
        expect = Expect(pme, state.info.bond_dim, state.info.bond_dim)
        expect.iprint = max(min(self.verbose - 4, 3), 0)
        expect.solve(True, state.center == 0)
        dmr = expect.get_1pdm_spatial(self.n_orbs)
        dm = np.array(dmr).copy()
        dmr.deallocate()

        return dm.transpose((1, 0))

    def make_rdm2(self, state=None, norb=None, nelec=None):
        '''
        Spin-traced two-particle density matrix in MO basis
        dm2[p,q,r,s] = \sum_{sigma,tau} <p_sigma^\dagger r_tau^\dagger s_tau q_sigma>
        '''

        if state is None:
            state = self.mps
            state.dot = 2
        
        # 2PDM MPO
        pmpo = PDM2MPOQC(self.hamil)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 2PDM
        pme = MovingEnvironment(pmpo, state, state, "2PDM")
        pme.init_environments(self.verbose >= 5)
        expect = Expect(pme, state.info.bond_dim, state.info.bond_dim)
        expect.iprint = max(min(self.verbose - 4, 3), 0)
        expect.solve(True, state.center == 0)
        dmr = expect.get_2pdm_spatial(self.n_orbs)
        dm = np.array(dmr, copy=True)

        return dm.transpose((0, 3, 1, 2))
    
    def make_rdm12(self, state=None, norb=None, nelec=None):
        dm1 = self.make_rdm1(state, norb, nelec)
        dm2 = self.make_rdm2(state, norb, nelec)
        return dm1, dm2


class DMRGCASCI(lib.StreamObject):
    """CASCI using DMRG as fcisolver"""

    def __init__(self, mf_or_mol, ncas, nelecas, ncore=None):
        self._casci = mcscf.CASCI(mf_or_mol, ncas, nelecas, ncore)
        self.mol = self._casci._scf.mol
        self._scf = self._casci._scf
        self.converged = False
        self.verbose = self.mol.verbose
        self.stdout = self.mol.stdout
        self.fcisolver = DMRG(self._scf)

        self.e_tot = 0
        self.e_cas = None

    def kernel(self, mo_coeff=None):

        if mo_coeff is None:
            mo_coeff = self._scf.mo_coeff
        
        self.mo_coeff = mo_coeff

        mc = self._casci
        mo_coeff_act = mo_coeff[:, mc.ncore:mc.ncore + mc.ncas].copy()
        if self.mol.groupname.lower() == 'c1':
            orb_sym = [0] * mo_coeff_act.shape[1]
        else:
            orb_sym = casci_symm.label_symmetry_(mc, mo_coeff_act).orbsym
        mc.mo_coeff = mo_coeff

        h1e, e_core = mc.get_h1cas()
        g2e = mc.get_h2cas()
        self.e_tot, self.ci = self.fcisolver.kernel(h1e, g2e,
            mc.ncas, mc.nelecas, e_core, orb_sym)

        self.converged = True
        self.e_corr = self.e_tot - self._scf.e_tot

        lib.logger.note(self, 'DMRGCI E = %.16g  E_corr = %.16g',
            self.e_tot, self.e_corr)
        
        return self.e_corr, self.ci
    
    def make_rdm1(self, state=None, norb=None, nelec=None):
        '''
        Spin-traced one-particle density matrix in AO basis.
        dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>
        '''
        
        mc = self._casci
        casdm1 = self.fcisolver.make_rdm1(state, norb, nelec)
        mocore = self.mo_coeff[:,:mc.ncore]
        mocas = self.mo_coeff[:, mc.ncore:mc.ncore + mc.ncas]
        dm1 = np.dot(mocore, mocore.T) * 2
        dm1 = dm1 + mocas @ casdm1 @ mocas.T
        return dm1


if __name__ == '__main__':

    from pyscf import gto, scf, fci

    mol = gto.M(
        atom='''
            O  0.000000  0.000000  0.000000
            H  0.758602  0.000000  0.504284
            H  0.758602  0.000000 -0.504284
        ''',
        basis='ccpvdz',
        verbose=4, symmetry=False, spin=0)

    # Mean Field
    mf = scf.RHF(mol).run()

    # DMRG FCI solver
    # mc = mcscf.CASCI(mf, 8, 8)
    # h1e, e_core = mc.get_h1cas()
    # g2e = mc.get_h2cas()
    # dmrg = DMRGCI(mf)
    # dmrg.dmrg_args['maxM'] = 2000
    # dmrg.dmrg_args['sweep_tol'] = 1E-16
    # ener, ci = dmrg.kernel(h1e, g2e, mc.ncas, mc.nelecas, e_core)
    # dm1, dm2 = dmrg.make_rdm12(ci, mc.ncas, mc.nelecas)
    # print(ener)

    # Reference FCI solver
    # xfci = fci.FCI(mol)
    # ener, ci = xfci.kernel(h1e, g2e, mc.ncas, mc.nelecas, ecore=e_core)
    # dm1x, dm2x = xfci.make_rdm12(ci, mc.ncas, mc.nelecas)
    # print(ener)
    # print(np.linalg.norm(dm1 - dm1x))
    # print(np.linalg.norm(dm2 - dm2x))
    
    # DMRG CASCI
    # myci = mcscf.CASCI(mf, 8, 8)
    # myci.fcisolver = DMRGCI(mf)
    # myci.fcisolver.dmrg_args['maxM'] = 2000
    # myci.fcisolver.dmrg_args['sweep_tol'] = 1E-16
    # myci = myci.run()
    # dm1 = myci.make_rdm1()

    # Reference CASCI
    # myci = mcscf.CASCI(mf, 8, 8)
    # myci.fcisolver.conv_tol = 1E-16
    # myci.run()
    # dm1x = myci.make_rdm1()
    # print(np.linalg.norm(dm1 - dm1x))

    # DMRG CASSCF
    mc = mcscf.CASSCF(mf, 6, 6)
    mc.fcisolver = DMRGCI(mf)
    mc.fcisolver.dmrg_args['maxM'] = 2000
    mc.fcisolver.dmrg_args['sweep_tol'] = 1E-16
    mc.run()

    # Reference CASSCF
    # mc = mcscf.CASSCF(mf, 6, 6)
    # mc.fcisolver.kernel()
    # mc.run()




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2021 Henrik R. Larsson <larsson@caltech.edu>
#  Copyright (C) 2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

""" Finite temperature DMRG

:author: Henrik R. Larsson, Sep 2021
        Based on zero temperature GFDMRG from Huanchen Zhai
"""

from block2 import SU2, SZ, Global, OpNamesSet, Threading, ThreadingTypes
from block2 import init_memory, release_memory, SiteIndex
from block2 import VectorUInt8, PointGroup, FCIDUMP, QCTypes, SeqTypes, OpNames, Random
from block2 import VectorUBond, VectorDouble, NoiseTypes, DecompositionTypes, EquationTypes
from block2 import OrbitalOrdering, VectorUInt16, TETypes
import time

def _init(SpinLabel):
    if SpinLabel == SU2:
        from block2.su2 import AncillaMPO, AncillaMPSInfo
        from block2.su2 import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
        from block2.su2 import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO, IdentityAddedMPO
        from block2.su2 import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
        from block2.su2 import VectorOpElement, LocalMPO
        from block2.su2 import TimeEvolution

        try:
            from block2.su2 import MPICommunicator
            hasMPI = True
        except ImportError:
            hasMPI = False
    else:
        from block2.sz import AncillaMPO, AncillaMPSInfo
        from block2.sz import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
        from block2.sz import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO, IdentityAddedMPO
        from block2.sz import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
        from block2.sz import VectorOpElement, LocalMPO
        from block2.sz import TimeEvolution

        try:
            from block2.sz import MPICommunicator
            hasMPI = True
        except ImportError:
            hasMPI = False
    import tools; tools.init(SpinLabel)
    from tools import saveMPStoDir, loadMPSfromDir
    import numpy as np
    import scipy
    from typing import List, Union, Tuple

    if hasMPI:
        MPI = MPICommunicator()
    else:
        class _MPI:
            rank = 0
        MPI = _MPI()



    _print = tools.getVerbosePrinter(MPI.rank == 0, flush=True)


    class FTDMRG:
        """
        Finite-temperature DMRG
        """

        def __init__(self, scratch='./nodex', memory=1 * 1E9, omp_threads=8, verbose=2,
                     print_statistics=True, mpi=None, delayed_contraction=True):
            """
            :param scratch: Used scratchdir
            :param memory: Stackmemory in bytes
            :param omp_threads: Number of omp threads
            :param verbose: Detail of print statements: 0 (quiet), 2 (per sweep), 3 (per iteration)
            :param print_statistics: Print memory statistics before/ after each call
            :param mpi: MPI class
            :param delayed_contraction: Use of it in MovingEnvironment (recommended if there are not multiple MEs)
            """
            self.fcidump = None
            self.hamil = None
            self.verbose = verbose
            self.scratch = scratch
            self.mpo_orig = None
            self.print_statistics = print_statistics
            self.mpi = mpi
            self.delayed_contraction = delayed_contraction
            self.idx = None # reorder
            self.ridx = None # inv reorder
            self.swap_pg = None # PointGroup.swap_XX; initialized in init_fcidump*
            self.orb_sym = None
            self.target = None
            self.n_physical_sites = None
            self.n_sites = None

            Random.rand_seed(0)
            isize = int(1e8) # hrl: typically <  200MB
            init_memory(isize=isize, dsize=int(memory), save_dir=scratch)
            Global.threading = Threading(
                ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global, omp_threads, omp_threads, 1)
            Global.threading.seq_type = SeqTypes.Tasked
            Global.frame.load_buffering = False
            Global.frame.save_buffering = False
            Global.frame.use_main_stack = False
            Global.frame.minimal_disk_usage = True

            if self.verbose >= 2:
                _print(Global.frame)
                _print(Global.threading)

            if mpi is not None:
                if SpinLabel == SU2:
                    from block2.su2 import ParallelRuleQC, ParallelRuleNPDMQC, ParallelRuleSiteQC
                    from block2.su2 import ParallelRuleSiteQC, ParallelRuleIdentity
                else:
                    from block2.sz import ParallelRuleQC, ParallelRuleNPDMQC
                    from block2.sz import ParallelRuleSiteQC, ParallelRuleIdentity
                self.prule = ParallelRuleQC(mpi)
                self.pdmrule = ParallelRuleNPDMQC(mpi)
                self.siterule = ParallelRuleSiteQC(mpi)
                self.identrule = ParallelRuleIdentity(mpi)
            else:
                self.prule = None
                self.pdmrule = None
                self.siterule = None
                self.identrule = None

        @staticmethod
        def fmt_size(i, suffix='B'):
            if i < 1000:
                return "%d %s" % (i, suffix)
            else:
                a = 1024
                for pf in "KMGTPEZY":
                    p = 2
                    for k in [10, 100, 1000]:
                        if i < k * a:
                            return "%%.%df %%s%%s" % p % (i / a, pf, suffix)
                        p -= 1
                    a *= 1024
            return "??? " + suffix

        def init_hamiltonian_fcidump(self, pointgroup: str, filename: str, idx=None):
            """Read integrals from FCIDUMP file

            :param pointgroup: Used point group  (c1, c2v,..)
            :param filename: FCIDUMP file
            :param idx: Optional orbital reordering index
            """
            assert self.fcidump is None
            self.fcidump = FCIDUMP()
            self.fcidump.read(filename)
            if idx is not None:
                self.fcidump.reorder(VectorUInt16(idx))
                self.idx = idx
                self.ridx = np.argsort(idx)
            self.swap_pg = getattr(PointGroup, "swap_" + pointgroup)
            self.orb_sym = VectorUInt8(
                map(self.swap_pg, self.fcidump.orb_sym))

            vacuum = SpinLabel(0)
            # hrl: twos should be 0 for thermal state; and n_elec = n_sites (2*n_sites_physical)
            self.target = SpinLabel(2 * self.fcidump.n_sites, 0, self.swap_pg(self.fcidump.isym))
            self.n_physical_sites = self.fcidump.n_sites
            self.n_sites = self.fcidump.n_sites * 2

            self.hamil = HamiltonianQC(
                vacuum, self.n_physical_sites, self.orb_sym, self.fcidump)

        def init_hamiltonian(self, n_elec: int, twos: int, isym: int, orb_sym: List[int],
                             e_core: float, h1e: np.ndarray, g2e: np.ndarray, tol=1E-13, idx=None,
                             save_fcidump=None):
            """ Initialize fcidump based on integrals

            :param pointgroup: Used point group  (c1, c2v,..)
            :param n_elec: Total number of electrons
            :param twos: 2S quantum number
            :param isym: Wavefunction point group symmetry
            :param orb_sym: Orbital symmetry in *molpro* notation (not PySCF XOR)
            :param e_core: Core energy
            :param h1e: One-electron integrals
            :param g2e: Two-electron integrals
            :param tol:  All integral values below tol will be set to zero
            :param idx: Optional orbital reordering index
            :param save_fcidump: Use this file to write an fcidump file (for later usage)
            """
            n_sites = h1e.shape[0]
            assert self.fcidump is None
            self.fcidump = FCIDUMP()
            self.swap_pg = getattr(PointGroup, "swap_" + pointgroup)
            if not isinstance(h1e, tuple):
                mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
                k = 0
                for i in range(0, n_sites):
                    for j in range(0, i + 1):
                        assert abs(h1e[i, j] - h1e[j, i]) < tol
                        mh1e[k] = h1e[i, j]
                        k += 1
                mg2e = g2e.ravel()
                mh1e[np.abs(mh1e) < tol] = 0.0
                mg2e[np.abs(mg2e) < tol] = 0.0
                self.fcidump.initialize_su2(
                    n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
            else:
                assert SpinLabel == SZ
                assert isinstance(h1e, tuple) and len(h1e) == 2
                assert isinstance(g2e, tuple) and len(g2e) == 3
                mh1e_a = np.zeros((n_sites * (n_sites + 1) // 2))
                mh1e_b = np.zeros((n_sites * (n_sites + 1) // 2))
                mh1e = (mh1e_a, mh1e_b)
                for xmh1e, xh1e in zip(mh1e, h1e):
                    k = 0
                    for i in range(0, n_sites):
                        for j in range(0, i + 1):
                            assert abs(xh1e[i, j] - xh1e[j, i]) < tol
                            xmh1e[k] = xh1e[i, j]
                            k += 1
                    xmh1e[np.abs(xmh1e) < tol] = 0.0
                mg2e = tuple(xg2e.ravel() for xg2e in g2e)
                for xmg2e in mg2e:
                    xmg2e[np.abs(xmg2e) < tol] = 0.0
                self.fcidump.initialize_sz(
                    n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
            assert not np.any(np.array(orb_sym) == 0), "orb_sym should be in molpro notation"
            self.fcidump.orb_sym = VectorUInt8(orb_sym)
            if idx is not None:
                self.fcidump.reorder(VectorUInt16(idx))
                self.idx = idx
                self.ridx = np.argsort(idx)
            self.orb_sym = VectorUInt8(map(self.swap_pg, self.fcidump.orb_sym))

            vacuum = SpinLabel(0)
            self.target = SpinLabel(2 * self.fcidump.n_sites, 0, self.swap_pg(self.fcidump.isym))
            self.n_physical_sites = self.fcidump.n_sites
            self.n_sites = self.fcidump.n_sites * 2

            self.hamil = HamiltonianQC(vacuum, self.n_sites, self.orb_sym, self.fcidump)

            if save_fcidump is not None:
                if self.mpi is None or self.mpi.rank == 0:
                    self.fcidump.orb_sym = VectorUInt8(orb_sym)
                    self.fcidump.write(save_fcidump)
                if self.mpi is not None:
                    self.mpi.barrier()

        def prepare_ground_state(self, mu: float,
                                 beta: float, dbeta: float,
                                 bond_dim: int,
                                 save_dir=None,
                                 tag="psi_t0", dot=2,
                                 n_sub_sweeps_init=6, n_sub_sweeps=2,
                                 normalize=True) -> Tuple[MPS,float]:
            """ Get the initial ground state by propagating a maximally entangled state until beta
            exp(-beta H) |max_entangled>
            Currently uses RK4 algorithm

            :param mu: Chemical potential added during the propagation (switched off afterward)
            :param beta: Inverse temperature
            :param dbeta: "time" step
            :param bond_dim: Max bond dimension for MPS
            :param save_dir: If not None, save final MPS to this directory
            :param tag: MPS tag
            :param dot: MPS dot (2 or 1)
            :param n_sub_sweeps_init: Initial number of sweeps for RK4 algorithm
            :param n_sub_sweeps:  Number of sweeps for RK4 algorithm
            :param normalize: Normalize mps during propagation
            :return: mps, energy - mu * <n>
            """
            assert self.fcidump is not None, "call init_hamiltonian first"
            mps_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum, self.target, self.hamil.basis)
            mps_info.tag = tag
            mps_info.set_thermal_limit()
            mps_info.save_mutable()
            mps_info.deallocate_mutable()

            assert dot == 2 or dot == 1
            mps = MPS(self.n_sites, self.n_sites - dot, dot)
            mps_info.load_mutable()
            mps.initialize(mps_info)
            mps.fill_thermal_limit()
            mps.canonicalize()

            mps.save_mutable()
            mps.deallocate()
            mps_info.deallocate_mutable()
            mps.save_data()


            # Propagation
            if self.verbose > 0:
                _print("##############################")
                _print("# Initial MPS propagation")
            n_steps = int( ( beta / 2) / dbeta + 1)  # propagate until beta/2 (ancillary system)
            dbeta = (beta / 2) / (n_steps + 1) # make sure we have the correct step
            if self.verbose > 0:
                _print("# # time evolution steps:", n_steps)
                _print("# used beta step:", dbeta)
            # MPO
            self.hamil.mu = mu
            mpo = MPOQC(self.hamil, QCTypes.Conventional)
            mpo = SimplifiedMPO(AncillaMPO(mpo), RuleQC(), True, True,
                                OpNamesSet((OpNames.R, OpNames.RD)))
            mpo = IdentityAddedMPO(mpo) # hrl: sometimes const_e causes trouble for AncillaMPO
            # TE
            me = MovingEnvironment(mpo, mps, mps, "ground_state_prep")
            if self.delayed_contraction:
                me.delayed_contraction = OpNamesSet.normal_ops()
            me.cached_contraction = True
            me.init_environments(False)
            te = TimeEvolution(me, VectorUBond([bond_dim]), TETypes.RK4, n_sub_sweeps_init)
            te.iprint = self.verbose
            te.normalize_mps = normalize
            te.solve(1, dbeta, mps.center == 0)
            te.n_sub_sweeps = n_sub_sweeps
            if n_steps != 1:
                # after the first beta step, use 2 sweeps (or 1 sweep) for each beta step
                te.solve(n_steps-1, dbeta, mps.center == 0)
            energy_minus_mu_n = te.energies[-1]

            mpo.deallocate()
            self.hamil.mu = 0
            if save_dir is not None:
                saveMPStoDir(mps, save_dir)
            if self.verbose > 0:
                _print("# done")
                _print("##############################")
            return mps, energy_minus_mu_n

        def optimize_mu(self, n_elec,
                        mu_init: float,
                        beta: float, dbeta: float,
                        bond_dim: int,
                        save_dir=None,
                        tol=1e-3, maxiter=10,
                        tag="psi_t0", dot=2,
                        n_sub_sweeps_init=6, n_sub_sweeps=2) -> Tuple[MPS, float]:
            """ Optimize chemical potential via conjugate gradient minimzation of <N>
            Currently only implemented for <N> = <Nalpha + Nbeta>; ie not for SZ

            :param n_elec: Total number of electrons
            :param mu_init: Initial value. This is important as the optimization landscape is not convex
            :param beta: Inverse temperature
            :param dbeta: "time" step
            :param bond_dim: Max bond dimension for MPS
            :param save_dir: If not None, save final MPS to this directory
            :param tag: MPS tag
            :param dot: MPS dot (2 or 1)
            :param n_sub_sweeps_init: Initial number of sweeps for RK4 algorithm
            :param n_sub_sweeps:  Number of sweeps for RK4 algorithm
            :return: mps, energy - mu * <n>
            :param tol: chemical potential optimization tolerance
            :param maxiter: Max iterations in minimizer
            :return: MPS and optimal chemical potential
            """
            # number operator
            assert self.fcidump is not None, "call init_hamiltonian first"
            assert self.hamil is not None, "call init_hamiltonian first"
            fcidump_n = FCIDUMP()
            n_sites = self.n_physical_sites
            h1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(n_sites):
                for j in range(i + 1):
                    if i == j:
                        h1e[k] = 1
                    k += 1
            g2e = np.zeros(n_sites**4)
            fcidump_n.initialize_su2(n_sites, self.fcidump.n_elec, self.fcidump.twos,
                                     self.fcidump.isym, 0.0, h1e, g2e)
            # number operator squared
            fcidump_nn = FCIDUMP()
            g2e = g2e.reshape([n_sites]*4)
            for i in range(n_sites):
                for j in range(n_sites):
                    g2e[i,i,j,j] = 2
            g2e = g2e.ravel()
            fcidump_nn.initialize_su2(n_sites, self.fcidump.n_elec, self.fcidump.twos,
                                     self.fcidump.isym, 0.0, h1e, g2e)
            del g2e, h1e
            hamil_n = HamiltonianQC(self.hamil.vacuum, self.n_physical_sites, self.orb_sym, fcidump_n)
            hamil_nn = HamiltonianQC(self.hamil.vacuum, self.n_physical_sites, self.orb_sym, fcidump_nn)
            mpo_n = MPOQC(hamil_n, QCTypes.Conventional)
            mpo_n = SimplifiedMPO(AncillaMPO(mpo_n), RuleQC(), True, True,
                                OpNamesSet((OpNames.R, OpNames.RD)))
            mpo_nn = MPOQC(hamil_nn, QCTypes.Conventional)
            mpo_nn = SimplifiedMPO(AncillaMPO(mpo_nn), RuleQC(), True, True,
                                  OpNamesSet((OpNames.R, OpNames.RD)))

            used_mps = [None] # mutable; changes in solve
            def solve(mu):
                # switch verbose off
                verbose = self.verbose
                if self.verbose < 3:
                    self.verbose = 0
                mps, energy_minus_mu_n = self.prepare_ground_state(mu, beta, dbeta, bond_dim, save_dir, tag, dot,
                                                n_sub_sweeps_init, n_sub_sweeps)
                used_mps[0] = mps
                self.verbose = verbose
                # expectation valeus
                me_n = MovingEnvironment(mpo_n, mps, mps, "me_n")
                me_nn = MovingEnvironment(mpo_nn, mps, mps, "me_nn")
                me_n.init_environments(False)
                me_nn.init_environments(False)

                erw_n = Expect(me_n, bond_dim, bond_dim).solve(False)
                erw_nn = Expect(me_nn, bond_dim, bond_dim).solve(False)
                derw_n = beta * (erw_nn - erw_n**2)
                diff = erw_n - n_elec
                fmu = diff**2  # optimize this
                gmu = 2 * diff * derw_n # gradent of fmu w.r.t mu
                E = energy_minus_mu_n + erw_n * mu
                if self.verbose > 1:
                    _print(f"mu = {mu:17.12f} E = {E:17.12f} <N> = {erw_n:17.12f} <N^2> = {erw_nn:17.12f} "
                      f" <dN/dmu>={derw_n:17.12f} | F[mu] = {fmu:17.12f} grad[mu]={gmu:17.12f}")
                return fmu, gmu

            # implement caching for solver
            mu_cache = {}
            def f(mu):
                mu = mu[0]
                if mu not in mu_cache:
                    mu_cache[mu] = solve(mu)
                return mu_cache[mu][0]

            def g(mu):
                mu = mu[0]
                if mu not in mu_cache:
                    mu_cache[mu] = solve(mu)
                return mu_cache[mu][1]


            if self.verbose > 1:
                _print("#~~~~~~~~~~~~~~~~~~~~")
                _print("# Start optimizing mu")
                opt_mu = scipy.optimize.minimize(f, mu_init, method="CG", jac=g, tol=tol,
                                                 options={'disp': self.verbose > 0, 'maxiter': maxiter})
            if self.verbose > 1:
                _print("# done. opt result=",opt_mu)
                _print("#~~~~~~~~~~~~~~~~~~~~")
            mpo_nn.deallocate()
            mpo_n.deallocate()
            fcidump_n.deallocate()
            fcidump_nn.deallocate()
            return used_mps[0], opt_mu.x[0]

        def __del__(self):
            if self.hamil is not None:
                self.hamil.deallocate()
            if self.fcidump is not None:
                self.fcidump.deallocate()
            if self.mpo_orig is not None:
                self.mpo_orig.deallocate()
            release_memory()

    return FTDMRG

FTDMRG_SZ = _init(SZ)
FTDMRG_SU2 = _init(SU2)




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2021 Henrik R. Larsson <larsson@caltech.edu>
#  Copyright (C) 2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

""" Finite temperature td-DMRG for Green's Function (time-domain => frequency domain).

:author: Henrik R. Larsson, Sep 2021
"""

from block2 import SU2, SZ, Global, OpNamesSet, Threading, ThreadingTypes
from block2 import init_memory, release_memory, SiteIndex
from block2 import VectorUInt8, PointGroup, FCIDUMP, QCTypes, SeqTypes, OpNames, Random
from block2 import VectorUBond, VectorDouble, NoiseTypes, DecompositionTypes, EquationTypes
from block2 import OrbitalOrdering, VectorUInt16, TETypes
import time

# Set spin-adapted or non-spin-adapted here
#SpinLabel = SU2
SpinLabel = SZ

if SpinLabel == SU2:
    from block2.su2 import AncillaMPO, AncillaMPSInfo
    from block2.su2 import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.su2 import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO, IdentityAddedMPO
    from block2.su2 import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.su2 import VectorOpElement, LocalMPO
    from block2.su2 import ComplexExpect, MultiMPS, TimeEvolution
    try:
        from block2.su2 import MPICommunicator
        hasMPI = True
    except ImportError:
        hasMPI = False
    from ft_dmrg import FTDMRG_SU2 as FTDMRG
else:
    from block2.sz import AncillaMPO, AncillaMPSInfo
    from block2.sz import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.sz import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO, IdentityAddedMPO
    from block2.sz import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.sz import VectorOpElement, LocalMPO
    from block2.sz import ComplexExpect, MultiMPS, TimeEvolution
    try:
        from block2.sz import MPICommunicator
        hasMPI = True
    except ImportError:
        hasMPI = False
    from ft_dmrg import FTDMRG_SZ as FTDMRG
import tools; tools.init(SpinLabel)
from tools import saveMPStoDir, loadMPSfromDir, changeCanonicalForm
import numpy as np
from typing import List, Tuple, Union

if hasMPI:
    MPI = MPICommunicator()
else:
    class _MPI:
        rank = 0
    MPI = _MPI()



_print = tools.getVerbosePrinter(MPI.rank == 0, flush=True)


class RT_GFDMRG(FTDMRG):
    """
    Finite temperature td-DMRG for Green's Function in time domain.
    """

    def greens_function(self, mps: MPS, E0: float,
                        tmax: float,
                        dt: float,
                        idxs: List[int],
                        bond_dim: int,
                        cps_bond_dims: List[int], cps_noises: List[float],
                        cps_conv_tol: float, cps_n_sweeps: float,
                        diag_only=False, alpha=True,
                        cutoff=1E-14,
                        max_solver_iter=20000,
                        max_solver_iter_off_diag=0,
                        occs=None, bias=1.0, mo_coeff=None,
                        callback=lambda i,j,t,gf:None,
                        n_sub_sweeps=2, n_sub_sweeps_init = 4, exp_tol = 1e-6) -> Tuple[np.ndarray, np.ndarray]:
        """ Solve for the Green's function in time-domain.
        GF_ij(t) = -i theta(t) <psi0| V_i' exp[i (H-E0) t] V_j |psi0>
        With V_i = a_i or a'_i. theta(t) is the step function (no time-reversal symm)
        Note the definition of the sign of frequency omega.
        The GF is solved via the correction MPS C = inv(H - E0 + omega + i eta) V |psi0>

        :param mps: Start state psi0
        :param E0: Hamiltonian shift, typically ground state energy
        :param tmax: Max. propagation time
        :param dt: time step
        :param idxs: GF orbital indices to compute GF_ij, with i,j in idxs
        :param bond_dim: Max bond dimension during propagation
        :param cps_bond_dims: Number of bond dimensions for each sweep for V |þsi0>
        :param cps_noises: Noises for each sweep for V |þsi0>
        :param cps_conv_tol:  Sweep convergence tolerance for V |þsi0>
        :param cps_n_sweeps: Number of sweeps for obtaining V |psi0>
        :param cutoff: Bond dimension cutoff for sweeps
        :param diag_only: Solve only diagonal of GF: GF_ii
        :param alpha: Creation/annihilation operator refers to alpha spin (otherwise: beta spin)
        :param max_solver_iter: Max. solver iterations for GF
        :param max_solver_iter_off_diag: Max. solver iteration for off diagonal terms (if 0, is set to max_solver_iter)
        :param occs: Optional occupation number vector for V|psi0> initialization
        :param bias: Optional occupation number bias for V|psi0> initialization
        :param mo_coeff: MPO is in MO basis but GF should be computed in AO basis
        :param callback: Callback function after each GF computation.
                        Called as callback(i,j,t,GF_ij(t), GF_ij(2t) * delta_ij)
        :param n_sub_sweeps: Number of sub sweeps for RK4 time evolution
        :param n_sub_sweeps_int:  Number of sub sweeps for RK4 time evolution during initial dt
        :param exp_tol: exp(M) |vec> solver tolerance
        :return: propagation times, GF matrix, diagonal of GF matrix(2*t) [assuming real-valued initial state]
        """
        ops = [None] * len(idxs)
        rkets = [None] * len(idxs)
        rmpos = [None] * len(idxs)

        if self.mpi is not None:
            self.mpi.barrier()

        if self.mpo_orig is None:
            mpo = MPOQC(self.hamil, QCTypes.Conventional)
            mpo = SimplifiedMPO(AncillaMPO(mpo), RuleQC(), True, True,
                                OpNamesSet((OpNames.R, OpNames.RD)))
            self.mpo_orig = mpo


        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO

        mpo = 1.0 * self.mpo_orig
        mpo.const_e -= E0

        #mpo.const_e = 0 # hrl: sometimes const_e causes trouble for AncillaMPO
        mpo = IdentityAddedMPO(mpo) # hrl: alternative

        if self.mpi is not None:
            mpo = ParallelMPO(mpo, self.prule)

        if self.print_statistics:
            _print('RT-GF MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            mps_info2 = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                self.target, self.hamil.basis)
            mps_info2.set_bond_dimension(bond_dim)
            _, mem2, disk = mpo.estimate_storage(mps_info2, 2)
            _print("RT-GF EST MAX MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info2.left_dims]))
            _print("RT-GF EST PEAK MEM = ", RT_GFDMRG.fmt_size(
                mem2), " SCRATCH = ", RT_GFDMRG.fmt_size(disk))
            mps_info2.deallocate_mutable()
            mps_info2.deallocate()

        impo = SimplifiedMPO(AncillaMPO(IdentityMPO(self.hamil)),
                             NoTransposeRule(RuleQC()), True, True, OpNamesSet((OpNames.R, OpNames.RD)))

        if self.mpi is not None:
            impo = ParallelMPO(impo, self.identrule)

        ############################################
        # Prepare creation/annihilation operators
        ############################################
        if mo_coeff is None:
            if self.ridx is not None:
                gidxs = self.ridx[np.array(idxs)]
            else:
                gidxs = idxs
        else:
            if self.idx is not None:
                mo_coeff = mo_coeff[:, self.idx]
            gidxs = list(range(self.n_sites))
            ops = [None] * self.n_sites
            _print('idxs = ', idxs, 'gidxs = ', gidxs)

        for ii, idx in enumerate(gidxs):
            if SpinLabel == SZ:
                ops[ii] = OpElement(OpNames.D, SiteIndex(
                    (idx, ), (0 if alpha else 1, )), SZ(-1, -1 if alpha else 1, self.orb_sym[idx]))
            else:
                ops[ii] = OpElement(OpNames.D, SiteIndex(
                    (idx, ), ()), SU2(-1, 1, self.orb_sym[idx]))

        ############################################
        # Solve V_i |psi0>
        ############################################
        for ii, idx in enumerate(idxs):
            if self.mpi is not None:
                self.mpi.barrier()
            if self.verbose >= 2:
                _print('>>> START Compression Site = %4d <<<' % idx)
            t = time.perf_counter()

            rket_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                       self.target + ops[ii].q_label, self.hamil.basis)
            rket_info.tag = 'DKET%d' % idx
            bond_dim = mps.info.bond_dim
            bond_dim = bond_dim if bond_dim != 0 else max([x.n_states_total for x in mps.info.left_dims])
            if occs is None:
                if self.verbose >= 2:
                    _print("Using FCI INIT MPS,bond_dim=", bond_dim)
                rket_info.set_bond_dimension(bond_dim)
            else:
                if self.verbose >= 2:
                    _print("Using occupation number INIT MPS; bond_dim=", bond_dim)
                rket_info.set_bond_dimension_using_occ(
                    bond_dim, VectorDouble(occs), bias=bias)
            rkets[ii] = MPS(self.n_sites, mps.center, 2)
            rkets[ii].initialize(rket_info)
            rkets[ii].random_canonicalize()

            rkets[ii].save_mutable()
            rkets[ii].deallocate()
            rket_info.save_mutable()
            rket_info.deallocate_mutable()

            if mo_coeff is None:
                # the mpo and gf are in the same basis
                # the mpo is SiteMPO
                rmpos[ii] = SimplifiedMPO(
                    AncillaMPO(SiteMPO(self.hamil, ops[ii])), NoTransposeRule(RuleQC()),
                    True, True, OpNamesSet((OpNames.R, OpNames.RD)))
            else:
                # the mpo is in mo basis and gf is in ao basis
                # the mpo is sum of SiteMPO (LocalMPO)
                ao_ops = VectorOpElement([None] * self.n_sites)
                for ix in range(self.n_physical_sites):
                    iix = ix * 2 # not ancilla sites
                    ao_ops[iix] = ops[iix] * mo_coeff[idx, ix]
                rmpos[ii] = SimplifiedMPO(
                    AncillaMPO(LocalMPO(self.hamil, ao_ops)), NoTransposeRule(RuleQC()),
                    True, True, OpNamesSet((OpNames.R, OpNames.RD)))

            if self.mpi is not None:
                rmpos[ii] = ParallelMPO(rmpos[ii], self.siterule)

            if len(cps_noises) == 1 and cps_noises[0] == 0:
                pme = None
            else:
                pme = MovingEnvironment(mpo, rkets[ii], rkets[ii], "PERT")
                pme.init_environments(False)
            rme = MovingEnvironment(rmpos[ii], rkets[ii], mps, "RHS")
            rme.init_environments(False)
            if self.delayed_contraction:
                if pme is not None:
                    pme.delayed_contraction = OpNamesSet.normal_ops()
                rme.delayed_contraction = OpNamesSet.normal_ops()

            # ME
            cps = Linear(pme, rme, VectorUBond(cps_bond_dims),
                         VectorUBond([mps.info.get_max_bond_dimension() + 100]),
                         VectorDouble(cps_noises))
            cps.noise_type = NoiseTypes.ReducedPerturbativeCollectedLowMem
            if pme is not None:
                cps.eq_type = EquationTypes.PerturbativeCompression
            cps.iprint = max(self.verbose - 1, 0)
            cps.cutoff = cutoff
            cps.solve(cps_n_sweeps, mps.center == 0, cps_conv_tol)


            if self.verbose >= 2:
                _print('>>> COMPLETE Compression Site = %4d | Time = %.2f <<<' %
                       (idx, time.perf_counter() - t))


        ############################################
        # Big GF LOOP
        ############################################
        n_steps = int(tmax/dt + 1)
        ts = np.linspace(0, tmax, n_steps) # times
        dt = ts[1]-ts[0]
        gf_mat = np.zeros((len(idxs), len(idxs), n_steps), dtype=complex)
        gf_mat2t = np.zeros((len(idxs), n_steps), dtype=complex) # diagonal
        ts2t = 2 * ts

        # for autocorrelation
        idMPO = SimplifiedMPO(AncillaMPO(IdentityMPO(self.hamil)), RuleQC(), True, True)
        for ii, idx in enumerate(idxs):

            #
            # make MPS complex-valued
            #
            mps = MultiMPS.make_complex(rkets[ii], "mps_t")
            mps_t0 = MultiMPS.make_complex(rkets[ii], "mps_t0")
            if mps.dot != 1: # change to 2dot
                mps.load_data()
                mps_t0.load_data()
                mps.canonical_form = 'M' + mps.canonical_form[1:]
                mps_t0.canonical_form = 'M' + mps_t0.canonical_form[1:]
                mps.dot = 2
                mps_t0.dot = 2
                mps.save_data()
                mps_t0.save_data()

            # MPO for autocorrelation
            idME = MovingEnvironment(idMPO, mps_t0, mps, "acorr")
            acorr = ComplexExpect(idME, bond_dim, bond_dim)

            me = MovingEnvironment(mpo, mps, mps, "TE")
            if self.delayed_contraction:
                me.delayed_contraction = OpNamesSet.normal_ops()
            me.cached_contraction = True
            me.init_environments()

            method = TETypes.RK4
            te = TimeEvolution(me, VectorUBond([bond_dim]), method, n_sub_sweeps_init)
            te.cutoff = 0  # for tiny systems, this is important
            te.iprint = 6  # ft.verbose
            te.normalize_mps = False

            for it, tt in enumerate(ts):

                if self.verbose >= 2:
                    _print('>>>   START  TD-GF TIME = %10.5f Site = %4d %4d <<<' %
                           (tt, idx, idx))
                t = time.perf_counter()

                if it != 0: # time zero: no propagation
                    te.solve(1, +1j * dt, mps.center == 0, tol=exp_tol)
                    te.n_sub_sweeps = n_sub_sweeps

                idME.init_environments()
                gf_mat[ii, ii, it] = acorr.solve(False) * -1j
                #
                # double time trick
                # acorr(2t) = <psi(t)*|psi(t)>: here: site of center wavefunction ("state averaged")
                # This only works for the diagonal of the GF matrix, though
                if mps.wfns[0].data.size == 0:
                    loaded = True
                    mps.load_tensor(mps.center)
                vec = mps.wfns[0].data + 1j * mps.wfns[1].data
                gf_mat2t[ii, it] = np.vdot(vec.conj(),vec) * -1j
                if loaded:
                    mps.unload_tensor(mps.center)
                callback(ii,ii,tt,gf_mat[ii,ii,it], gf_mat2t[ii,it])
                #

                dmain, dseco, imain, iseco = Global.frame.peak_used_memory

                if self.verbose >= 1:
                    rgf, igf = gf_mat[ii,ii,it].real, gf_mat[ii,ii,it].imag
                    _print("=== TD-GF (%4d%4d | TIME = %10.5f ) = RE %20.15f + IM %20.15f === T = %7.2f" %
                           (idx, idx, tt, rgf, igf, time.perf_counter() - t))

                if self.verbose >= 2:
                    _print('>>> COMPLETE TD-GF TIME = %10.5f Site = %4d %4d | Time = %.2f <<<' %
                           (tt, idx, idx, time.perf_counter() - t))

                if diag_only:
                    continue

                for jj, idx2 in enumerate(idxs):
                    if jj > ii and rkets[jj].info.target == rkets[ii].info.target:

                        if self.verbose >= 2:
                            _print('>>>   START  TD-GF TIME = %10.5f Site = %4d %4d <<<' % (
                                tt, idx2, idx))
                        t = time.perf_counter()
                        if rkets[jj].center != mps.center:
                            # TODO: don't do this on rkets[jj] directly
                            changeCanonicalForm(rkets[jj], self.identrule)
                        mps_t0j = MultiMPS.make_complex(rkets[jj], "mps_t0j")
                        assert mps_t0j.center == mps.center, f"\nmps[j] form: {mps_t0j.canonical_form} \n" \
                                                                 f" mps[i] form: {mps.canonical_form}"

                        idMEj = MovingEnvironment(idMPO, mps_t0j, mps, "acorr_j")
                        idMEj.init_environments()
                        acorrj = ComplexExpect(idMEj, bond_dim, bond_dim)
                        gf_mat[jj, ii, it] = gf_mat[ii, jj, it] = acorrj.solve(False) * -1j

                        callback(ii,jj,tt,gf_mat[ii,jj,it],0.0)

                        if self.verbose >= 1:
                            rgf, igf = gf_mat[ii, jj, it].real, gf_mat[ii, jj, it].imag
                            _print("=== TD-GF (%4d%4d | TIME = %10.5f ) = RE %20.15f + IM %20.15f === T = %7.2f" %
                                   (idx2, idx, tt, rgf, igf, time.perf_counter() - t))
                        if self.verbose >= 2:
                            _print('>>> COMPLETE GF TIME = %10.5f Site = %4d %4d | Time = %.2f <<<' %
                                   (tt, idx2, idx, time.perf_counter() - t))
        mps.save_data()
        idMPO.deallocate()
        mps.info.deallocate()

        if self.print_statistics:
            _print("TD-GF PEAK MEM USAGE:",
                   "DMEM = ", RT_GFDMRG.fmt_size(dmain + dseco),
                   "(%.0f%%)" % (dmain * 100 / (dmain + dseco)),
                   "IMEM = ", RT_GFDMRG.fmt_size(imain + iseco),
                   "(%.0f%%)" % (imain * 100 / (imain + iseco)))

        return ts, gf_mat, gf_mat2t

    def fourier_transform_gf(self,ts: np.ndarray, gf:np.ndarray, eta:float)\
            -> Tuple[np.ndarray, np.ndarray]:
        """  Fourier transform the GF
         (assuming that the time-date of the gf-tensor is in t the last dimension)

        :param ts: Propagation times
        :param gf: Greens function tensor: i x j x ts
        :param eta: Regularization
        :returns: omegas, gf in frequency domain
        """
        assert np.allclose(np.linspace(0,ts[-1],len(ts)), ts), "assume evenly spaced ts"
        dt = ts[1] - ts[0]
        omegas = np.fft.fftshift(np.fft.fftfreq(len(ts), dt)) * 2 * np.pi
        gf_freq = np.fft.fftshift(np.fft.fft(gf * np.exp(-eta*ts)),axes=-1) * dt
        gf_freq.real *= -1 # make consistent with GF definition
        return omegas, gf_freq


if __name__ == "__main__":

    # parameters
    n_threads = 2
    point_group = 'c1'
    scratch = '/tmp/block2'
    load_dir = None

    import os
    if MPI.rank == 0:
        tools.mkDir(scratch)
    if hasMPI:
        MPI.barrier()
    os.environ['TMPDIR'] = scratch
    E0 = -2.1379703474141984 #H4
    E0 = -1.568351864513 # H3


    MAX_M = 50
    cps_bond_dims=[MAX_M] # Bond dimension of \hat a_i |psi> *and* |psi> (in time evolution)
    cps_noises=[0]
    cps_tol=1E-7
    cps_n_sweeps=30

    beta = 80 # inverse temperature
    dbeta = 0.5 # "time" step
    mu = -0.026282794560 # Chemical potential for initial state preparation

    #################################################
    # Prepare initial state
    #################################################

    dmrg = RT_GFDMRG(scratch=scratch, memory=4e9,
                  verbose=3, omp_threads=n_threads)
    dmrg.init_hamiltonian_fcidump(point_group, "fcidump")
    #mps, mu = dmrg.optimize_mu(dmrg.fcidump.n_elec,mu, beta, dbeta, MAX_M)
    mps = dmrg.prepare_ground_state(mu, beta, dbeta, MAX_M)[0]


    tEnd = np.pi / 0.01
    dt = 0.1 # 1 / max(E)
    idxs = [0] # Calc S_ii
    alpha = True # alpha or beta spin
    fOut = open("test_ft_tddmrg_gf.dat","w")
    fOut.write("# t  Re(gf)  Im(gf)\n")
    fOut2 = open("test_ft_tddmrg_gf_2t.dat","w")
    fOut2.write("# t  Re(gf)  Im(gf)\n")
    def callback(i,j,t,gf, gf2t):
        print("-----",t,":",gf)
        fOut.write(f"{t:16.7f}   {gf.real:16.7f}  {gf.imag:16.7f}\n")
        fOut.flush()
        fOut2.write(f"{2*t:16.7f}   {gf2t.real:16.7f}  {gf2t.imag:16.7f}\n")
        fOut2.flush()
    ts, gf ,gf2t = dmrg.greens_function(mps, E0, tEnd, dt, idxs,
                          MAX_M,
                          cps_bond_dims, cps_noises, cps_tol, cps_n_sweeps,
                          diag_only=False,
                          alpha=alpha, max_solver_iter_off_diag=0,
                          callback=callback)
    eta = 0.005
    omegas, gf_freq = dmrg.fourier_transform_gf(ts, gf, eta)
    gf_freq = gf_freq.ravel()

    fOut.close()
    fOut2.close()
    fOut = open("test_ft_tddmrg_gf_ft.dat","w")
    fOut.write("# omega  Re(gf)  Im(gf)\n")
    for i in range(len(omegas)):
        fOut.write(f"{omegas[i]:16.7f}   {gf_freq[i].real:16.7f}  {gf_freq[i].imag:16.7f}\n")
    fOut.close()
    fOut = open("test_ft_tddmrg_gf_ft2.dat","w")
    omegas, gf_freq = dmrg.fourier_transform_gf(2*ts, gf2t, eta)
    gf_freq = gf_freq.ravel()
    fOut.write("# omega  Re(gf)  Im(gf)\n")
    for i in range(len(omegas)):
        fOut.write(f"{omegas[i]:16.7f}   {gf_freq[i].real:16.7f}  {gf_freq[i].imag:16.7f}\n")
    fOut.close()




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
MPS/MPO format transform between block2 and pyblock2.
"""

import numpy as np
from .core import MPS, MPO, Tensor, SubTensor
from block2 import OpTypes, QCTypes, SZ
from block2.sz import StateInfo, MPOQC


class TensorTools:
    @staticmethod
    def from_block2_fused(bspmat, l, r, lr, clr):
        """Translate block2 rank2 left-fused right-boundary SparseMatrix to pyblock2 rank2 tensor."""
        blocks = []
        for i in range(bspmat.info.n):
            ql = bspmat.info.quanta[i].get_bra(bspmat.info.delta_quantum)
            ib = lr.find_state(ql)
            bbed = clr.n if ib == lr.n - 1 else clr.n_states[ib + 1]
            pmat = np.array(bspmat[i]).flatten()
            ip = 0
            for bb in range(clr.n_states[ib], bbed):
                ibba = clr.quanta[bb].data >> 16
                ibbb = clr.quanta[bb].data & 0xFFFF
                ql, nl = l.quanta[ibba], l.n_states[ibba]
                qr, nr = r.quanta[ibbb], r.n_states[ibbb]
                rmat = pmat[ip: ip + nl * nr].reshape((nl, nr))
                blocks.append(
                    SubTensor(q_labels=(ql, qr), reduced=rmat.copy()))
                ip += nl * nr
            assert ip == pmat.shape[0]
        return Tensor(blocks=blocks)

    @staticmethod
    def from_block2_no_fused(bspmat):
        """Translate block2 rank2 unfused SparseMatrix to pyblock2 rank2 tensor."""
        blocks = []
        for i in range(bspmat.info.n):
            ql = bspmat.info.quanta[i].get_bra(bspmat.info.delta_quantum)
            qr = bspmat.info.quanta[i].get_ket()
            if bspmat.info.is_wavefunction:
                qr = -qr
            pmat = np.array(bspmat[i])
            blocks.append(
                SubTensor(q_labels=(ql, qr), reduced=pmat.copy()))
        return Tensor(blocks=blocks)

    @staticmethod
    def from_block2_left_fused(bspmat, l, m, lm, clm):
        """Translate block2 rank2 left-fused SparseMatrix to pyblock2 rank3 tensor."""
        blocks = []
        for i in range(bspmat.info.n):
            qlm = bspmat.info.quanta[i].get_bra(bspmat.info.delta_quantum)
            qr = bspmat.info.quanta[i].get_ket()
            if bspmat.info.is_wavefunction:
                qr = -qr
            ib = lm.find_state(qlm)
            bbed = clm.n if ib == lm.n - 1 else clm.n_states[ib + 1]
            pmat = np.array(bspmat[i])
            nr = pmat.shape[1]
            ip = 0
            for bb in range(clm.n_states[ib], bbed):
                ibba = clm.quanta[bb].data >> 16
                ibbb = clm.quanta[bb].data & 0xFFFF
                ql, nl = l.quanta[ibba], l.n_states[ibba]
                qm, nm = m.quanta[ibbb], m.n_states[ibbb]
                rmat = pmat[ip: ip + nl * nm, :].reshape((nl, nm, nr))
                blocks.append(
                    SubTensor(q_labels=(ql, qm, qr), reduced=rmat.copy()))
                ip += nl * nm
            assert ip == pmat.shape[0]
        return Tensor(blocks=blocks)

    @staticmethod
    def from_block2_right_fused(bspmat, m, r, mr, cmr):
        """Translate block2 rank2 right-fused SparseMatrix to pyblock2 rank3 tensor."""
        blocks = []
        for i in range(bspmat.info.n):
            ql = bspmat.info.quanta[i].get_bra(bspmat.info.delta_quantum)
            qmr = bspmat.info.quanta[i].get_ket()
            if bspmat.info.is_wavefunction:
                qmr = -qmr
            ik = mr.find_state(qmr)
            kked = cmr.n if ik == mr.n - 1 else cmr.n_states[ik + 1]
            pmat = np.array(bspmat[i])
            nl = pmat.shape[0]
            ip = 0
            for kk in range(cmr.n_states[ik], kked):
                ikka = cmr.quanta[kk].data >> 16
                ikkb = cmr.quanta[kk].data & 0xFFFF
                qm, nm = m.quanta[ikka], m.n_states[ikka]
                qr, nr = r.quanta[ikkb], r.n_states[ikkb]
                rmat = pmat[:, ip:ip + nm * nr].reshape((nl, nm, nr))
                blocks.append(
                    SubTensor(q_labels=(ql, qm, qr), reduced=rmat.copy()))
                ip += nm * nr
            assert ip == pmat.shape[1]
        return Tensor(blocks=blocks)

    @staticmethod
    def from_block2_left_and_right_fused(bspmat, l, ma, mb, r, lm, clm, mr, cmr):
        """Translate block2 rank2 left-and-right-fused SparseMatrix to pyblock2 rank4 tensor."""
        blocks = []
        for i in range(bspmat.info.n):
            qlm = bspmat.info.quanta[i].get_bra(bspmat.info.delta_quantum)
            qmr = bspmat.info.quanta[i].get_ket()
            if bspmat.info.is_wavefunction:
                qmr = -qmr
            ib = lm.find_state(qlm)
            bbed = clm.n if ib == lm.n - 1 else clm.n_states[ib + 1]
            ik = mr.find_state(qmr)
            kked = cmr.n if ik == mr.n - 1 else cmr.n_states[ik + 1]
            pmat = np.array(bspmat[i])
            ipl = 0
            for bb in range(clm.n_states[ib], bbed):
                ibba = clm.quanta[bb].data >> 16
                ibbb = clm.quanta[bb].data & 0xFFFF
                ql, nl = l.quanta[ibba], l.n_states[ibba]
                qma, nma = ma.quanta[ibbb], ma.n_states[ibbb]
                ipr = 0
                for kk in range(cmr.n_states[ik], kked):
                    ikka = cmr.quanta[kk].data >> 16
                    ikkb = cmr.quanta[kk].data & 0xFFFF
                    qmb, nmb = mb.quanta[ikka], mb.n_states[ikka]
                    qr, nr = r.quanta[ikkb], r.n_states[ikkb]
                    rmat = pmat[ipl: ipl + nl * nma, ipr: ipr
                                + nmb * nr].reshape((nl, nma, nmb, nr))
                    blocks.append(
                        SubTensor(q_labels=(ql, qma, qmb, qr), reduced=rmat.copy()))
                    ipr += nmb * nr
                assert ipr == pmat.shape[1]
                ipl += nl * nma
            assert ipl == pmat.shape[0]
        return Tensor(blocks=blocks)


class MPSTools:
    @staticmethod
    def from_block2(bmps):
        """Translate block2 MPS to pyblock2 MPS."""
        tensors = [None] * bmps.n_sites
        for i in range(0, bmps.n_sites):
            if bmps.tensors[i] is None:
                continue
            if (i == 0 and i < bmps.center) or (
                    i == bmps.n_sites - 1
                    and i >= bmps.center + bmps.dot) or (
                    i == 0 and i == bmps.center and bmps.dot == 1) or (
                    i == 0 and i == bmps.center
                    and i == bmps.n_sites - 2 and bmps.dot == 2):
                bmps.load_tensor(i)
                tensors[i] = TensorTools.from_block2_no_fused(bmps.tensors[i])
                bmps.unload_tensor(i)
            elif i < bmps.center or (
                    i == bmps.center and i == bmps.n_sites - 2 and
                    bmps.dot == 2) or (
                    i == bmps.center and bmps.dot == 1):
                bmps.info.load_left_dims(i)
                l = bmps.info.left_dims[i]
                m = bmps.info.basis[i]
                lm = StateInfo.tensor_product_ref(
                    l, m, bmps.info.left_dims_fci[i + 1])
                clm = StateInfo.get_connection_info(l, m, lm)
                bmps.load_tensor(i)
                if i == bmps.n_sites - 1 and i == bmps.center and bmps.dot == 1:
                    if bmps.tensors[i].info.n == 1 and \
                            bmps.tensors[i].info.quanta[0].get_ket() == -bmps.target:
                        tensors[i] = TensorTools.from_block2_fused(
                            bmps.tensors[i], l, m, lm, clm)
                    else:
                        tensors[i] = TensorTools.from_block2_no_fused(
                            bmps.tensors[i])
                else:
                    tensors[i] = TensorTools.from_block2_left_fused(
                        bmps.tensors[i], l, m, lm, clm)
                bmps.unload_tensor(i)
                clm.deallocate()
                lm.deallocate()
                l.deallocate()
            elif i >= bmps.center + bmps.dot or (
                    i == bmps.center and i == 0 and bmps.dot == 2):
                if i >= bmps.center + bmps.dot:
                    bmps.info.load_right_dims(i + 1)
                    m = bmps.info.basis[i]
                    r = bmps.info.right_dims[i + 1]
                    mr = StateInfo.tensor_product_ref(
                        m, r, bmps.info.right_dims_fci[i])
                else:
                    bmps.info.load_right_dims(i + 2)
                    m = bmps.info.basis[i + 1]
                    r = bmps.info.right_dims[i + 2]
                    mr = StateInfo.tensor_product_ref(
                        m, r, bmps.info.right_dims_fci[i + 1])
                cmr = StateInfo.get_connection_info(m, r, mr)
                bmps.load_tensor(i)
                tensors[i] = TensorTools.from_block2_right_fused(
                    bmps.tensors[i], m, r, mr, cmr)
                bmps.unload_tensor(i)
                cmr.deallocate()
                mr.deallocate()
                r.deallocate()
            elif i == bmps.center and i != 0 and i != bmps.n_sites - 2 and bmps.dot == 2:
                bmps.info.load_left_dims(i)
                bmps.info.load_right_dims(i + 2)
                l = bmps.info.left_dims[i]
                ma = bmps.info.basis[i]
                mb = bmps.info.basis[i + 1]
                r = bmps.info.right_dims[i + 2]
                lm = StateInfo.tensor_product_ref(
                    l, ma, bmps.info.left_dims_fci[i + 1])
                mr = StateInfo.tensor_product_ref(
                    mb, r, bmps.info.right_dims_fci[i + 1])
                clm = StateInfo.get_connection_info(l, m, lm)
                cmr = StateInfo.get_connection_info(m, r, mr)
                bmps.load_tensor(i)
                tensors[i] = TensorTools.from_block2_left_and_right_fused(
                    bmps.tensors[i], l, ma, mb, r, lm, clm, mr, cmr)
                bmps.unload_tensor(i)
                cmr.deallocate()
                clm.deallocate()
                mr.deallocate()
                lm.deallocate()
                r.deallocate()
                l.deallocate()
            else:
                assert False
        if bmps.center != bmps.n_sites - 1:
            for block in tensors[bmps.center].blocks:
                block.q_labels = block.q_labels[:-1] + \
                    (bmps.info.target - block.q_labels[-1], )
        for i in range(bmps.center + bmps.dot, bmps.n_sites):
            for block in tensors[i].blocks:
                if block.rank == 3:
                    block.q_labels = (bmps.info.target - block.q_labels[0],
                                      block.q_labels[1], bmps.info.target - block.q_labels[2])
                elif block.rank == 2:
                    block.q_labels = (bmps.info.target - block.q_labels[0],
                                      block.q_labels[1])
                else:
                    assert False
        return MPS(tensors=tensors)


class MPOTools:
    @staticmethod
    def from_block2(bmpo):
        """Translate block2 (un-simplified) MPO to pyblock2 MPO."""
        assert bmpo.schemer is None
        if isinstance(bmpo, MPOQC):
            assert bmpo.mode == QCTypes.NC or bmpo.mode == QCTypes.CN
        tensors = [None] * bmpo.n_sites
        # tranlate operator name symbols to quantum labels
        idx_mps, idx_qss, idx_imps = [], [], []
        for i in range(0, bmpo.n_sites - 1):
            lidx_mp = {}
            lidx_qs = [op.q_label for op in bmpo.left_operator_names[i].data]
            for ip, q in enumerate(lidx_qs):
                if q not in lidx_mp:
                    lidx_mp[q] = []
                lidx_mp[q].append(ip)
            limp = {iv: iiv for _, v in lidx_mp.items()
                    for iiv, iv in enumerate(v)}
            idx_mps.append(lidx_mp)
            idx_qss.append(lidx_qs)
            idx_imps.append(limp)
        for i in range(0, bmpo.n_sites):
            assert bmpo.tensors[i].lmat == bmpo.tensors[i].rmat
            mat = bmpo.tensors[i].lmat
            ops = bmpo.tensors[i].ops
            map_blocks = {}
            if i == 0:
                for k, expr in enumerate(mat.data):
                    if expr.get_type() == OpTypes.Zero:
                        continue
                    elif expr.get_type() == OpTypes.Elem:
                        spmat = ops[expr.abs()]
                        if spmat.factor == 0 or spmat.info.n == 0:
                            continue
                        qr = idx_qss[i][k]
                        nr = len(idx_mps[i][qr])
                        ir = idx_imps[i][k]
                        for p in range(spmat.info.n):
                            qu = spmat.info.quanta[p].get_bra(
                                spmat.info.delta_quantum)
                            qd = spmat.info.quanta[p].get_ket()
                            nu = spmat.info.n_states_bra[p]
                            nd = spmat.info.n_states_ket[p]
                            qx = (qu, qd, qr)
                            if qx not in map_blocks:
                                map_blocks[qx] = SubTensor(
                                    q_labels=qx, reduced=np.zeros((nu, nd, nr)))
                            map_blocks[qx].reduced[:, :, ir] += expr.factor * \
                                spmat.factor * np.array(spmat[p])
                    else:
                        assert False
            elif i == bmpo.n_sites - 1:
                for k, expr in enumerate(mat.data):
                    if expr.get_type() == OpTypes.Zero:
                        continue
                    elif expr.get_type() == OpTypes.Elem:
                        spmat = ops[expr.abs()]
                        if spmat.factor == 0 or spmat.info.n == 0:
                            continue
                        ql = idx_qss[i - 1][k]
                        nl = len(idx_mps[i - 1][ql])
                        il = idx_imps[i - 1][k]
                        for p in range(spmat.info.n):
                            qu = spmat.info.quanta[p].get_bra(
                                spmat.info.delta_quantum)
                            qd = spmat.info.quanta[p].get_ket()
                            nu = spmat.info.n_states_bra[p]
                            nd = spmat.info.n_states_ket[p]
                            qx = (ql, qu, qd)
                            if qx not in map_blocks:
                                map_blocks[qx] = SubTensor(
                                    q_labels=qx, reduced=np.zeros((nl, nu, nd)))
                            map_blocks[qx].reduced[il, :, :] += expr.factor * \
                                spmat.factor * np.array(spmat[p])
                    else:
                        assert False
            else:
                for (j, k), expr in zip(mat.indices, mat.data):
                    if expr.get_type() == OpTypes.Zero:
                        continue
                    elif expr.get_type() == OpTypes.Elem:
                        spmat = ops[expr.abs()]
                        if spmat.factor == 0 or spmat.info.n == 0:
                            continue
                        ql, qr = idx_qss[i - 1][j], idx_qss[i][k]
                        nl, nr = len(idx_mps[i - 1][ql]
                                     ), len(idx_mps[i][qr])
                        il, ir = idx_imps[i - 1][j], idx_imps[i][k]
                        for p in range(spmat.info.n):
                            qu = spmat.info.quanta[p].get_bra(
                                spmat.info.delta_quantum)
                            qd = spmat.info.quanta[p].get_ket()
                            nu = spmat.info.n_states_bra[p]
                            nd = spmat.info.n_states_ket[p]
                            qx = (ql, qu, qd, qr)
                            if np.linalg.norm(np.array(spmat[p])) == 0:
                                continue
                            if qx not in map_blocks:
                                map_blocks[qx] = SubTensor(
                                    q_labels=qx, reduced=np.zeros((nl, nu, nd, nr)))
                            map_blocks[qx].reduced[il, :, :,
                                                   ir] += expr.factor * spmat.factor * np.array(spmat[p])
                    else:
                        assert False
            tensors[i] = Tensor(blocks=list(map_blocks.values()))
        return MPO(tensors=tensors, const_e=bmpo.const_e)




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
Finite-Temperature (FT)-DMRG using pyscf and block2.

Author: Huanchen Zhai, May 14, 2020
Revised:     added sz, May 18, 2020
"""

from block2 import SU2, SZ
from block2 import Random, FCIDUMP, QCTypes, SeqTypes, TETypes, OpNamesSet, OpNames
from block2 import VectorUInt8, VectorUBond, VectorDouble, PointGroup
from block2 import init_memory, release_memory, Threading, ThreadingTypes, Global
import time
import numpy as np

# Set spin-adapted or non-spin-adapted here
# SpinLabel = SU2
SpinLabel = SZ

if SpinLabel == SU2:
    from block2.su2 import HamiltonianQC, AncillaMPSInfo, MPS
    from block2.su2 import AncillaMPO, PDM1MPOQC, NPC1MPOQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.su2 import MovingEnvironment, TimeEvolution, Expect, IdentityMPO, Linear
else:
    from block2.sz import HamiltonianQC, AncillaMPSInfo, MPS
    from block2.sz import AncillaMPO, PDM1MPOQC, PDM2MPOQC, NPC1MPOQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.sz import MovingEnvironment, TimeEvolution, Expect, IdentityMPO, Linear


class FTDMRGError(Exception):
    pass


class FTDMRG:
    """
    Finite-temperature DMRG for molecules.
    """

    def __init__(self, scratch='./nodex', memory=1 * 1E9, omp_threads=2, verbose=2):

        Random.rand_seed(0)
        init_memory(isize=int(memory * 0.1),
                    dsize=int(memory * 0.9), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global, omp_threads, omp_threads, 1)
        Global.threading.seq_type = SeqTypes.Tasked
        Global.frame.use_main_stack = False
        self.fcidump = None
        self.hamil = None
        self.verbose = verbose

    def init_hamiltonian_fcidump(self, pg, filename):
        assert self.fcidump is None
        self.fcidump = FCIDUMP()
        self.fcidump.read(filename)
        self.orb_sym = VectorUInt8(
            map(PointGroup.swap_d2h, self.fcidump.orb_sym))
        n_elec = self.fcidump.n_sites * 2

        vacuum = SpinLabel(0)
        self.target = SpinLabel(n_elec, 0, # twos is 0 for a thermal state; n_elec = n_sites
                                PointGroup.swap_d2h(self.fcidump.isym))
        self.n_physical_sites = self.fcidump.n_sites
        self.n_sites = self.fcidump.n_sites * 2

        self.hamil = HamiltonianQC(
            vacuum, self.n_physical_sites, self.orb_sym, self.fcidump)
        assert pg in ["d2h", "c1"]

    def init_hamiltonian(self, pg, n_sites, twos, isym, orb_sym, e_core, h1e, g2e, tol=1E-13, save_fcidump=None):
        assert self.fcidump is None
        self.fcidump = FCIDUMP()
        n_elec = n_sites * 2
        if not isinstance(h1e, tuple):
            mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    assert abs(h1e[i, j] - h1e[j, i]) < tol
                    mh1e[k] = h1e[i, j]
                    k += 1
            mg2e = g2e.flatten()
            mh1e[np.abs(mh1e) < tol] = 0.0
            mg2e[np.abs(mg2e) < tol] = 0.0
            self.fcidump.initialize_su2(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        else:
            assert SpinLabel == SZ
            assert isinstance(h1e, tuple) and len(h1e) == 2
            assert isinstance(g2e, tuple) and len(g2e) == 3
            mh1e_a = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e_b = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e = (mh1e_a, mh1e_b)
            for xmh1e, xh1e in zip(mh1e, h1e):
                k = 0
                for i in range(0, n_sites):
                    for j in range(0, i + 1):
                        assert abs(xh1e[i, j] - xh1e[j, i]) < tol
                        xmh1e[k] = xh1e[i, j]
                        k += 1
                xmh1e[np.abs(xmh1e) < tol] = 0.0
            mg2e = tuple(xg2e.flatten() for xg2e in g2e)
            for xmg2e in mg2e:
                xmg2e[np.abs(xmg2e) < tol] = 0.0
            self.fcidump.initialize_sz(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        self.orb_sym = VectorUInt8(map(PointGroup.swap_d2h, orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(n_elec, 0, PointGroup.swap_d2h(isym))
        self.n_physical_sites = n_sites
        self.n_sites = n_sites * 2

        self.hamil = HamiltonianQC(
            vacuum, self.n_physical_sites, self.orb_sym, self.fcidump)

        if save_fcidump is not None:
            self.fcidump.orb_sym = VectorUInt8(orb_sym)
            self.fcidump.write(save_fcidump)
        assert pg in ["d2h", "c1"]

    def generate_initial_mps(self, bond_dim):
        if self.verbose >= 2:
            print('>>> START generate initial mps <<<')
        t = time.perf_counter()
        assert self.hamil is not None

        # Ancilla MPSInfo (thermal)
        mps_info_thermal = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                          self.target, self.hamil.basis)
        mps_info_thermal.set_thermal_limit()
        mps_info_thermal.tag = "INIT"
        mps_info_thermal.save_mutable()
        mps_info_thermal.deallocate_mutable()

        if self.verbose >= 2:
            print("left dims = ", [
                p.n_states_total for p in mps_info_thermal.left_dims])
            print("right dims = ", [
                p.n_states_total for p in mps_info_thermal.right_dims])

        # Ancilla MPS (thermal)
        mps_thermal = MPS(self.n_sites, self.n_sites - 2, 2)
        mps_info_thermal.load_mutable()
        mps_thermal.initialize(mps_info_thermal)
        mps_thermal.fill_thermal_limit()
        mps_thermal.canonicalize()

        mps_thermal.save_mutable()
        mps_thermal.deallocate()
        mps_info_thermal.deallocate_mutable()

        mps_thermal.save_data()
        mps_info_thermal.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE generate initial mps | Time = %.2f <<<' %
                  (time.perf_counter() - t))

    def imaginary_time_evolution(self, n_steps, beta_step, mu, bond_dims,
                                 method=TETypes.RK4, n_sub_sweeps=4, cont=False):
        if self.verbose >= 2:
            print('>>> START imaginary time evolution <<<')
        t = time.perf_counter()

        self.hamil.mu = mu

        # Ancilla MPSInfo (initial)
        mps_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                  self.target, self.hamil.basis)
        mps_info.tag = "INIT" if not cont else "FINAL"
        mps_info.load_mutable()

        # Ancilla MPS (initial)
        mps = MPS(mps_info)
        mps.load_data()
        mps.load_mutable()

        # MPS/MPSInfo save mutable
        if not cont:
            mps_info.tag = "FINAL"
            mps_info.save_mutable()
            mps.save_mutable()
        mps.deallocate()
        mps_info.deallocate_mutable()

        # MPO
        tx = time.perf_counter()
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(AncillaMPO(mpo), RuleQC(), True, True,
                            OpNamesSet((OpNames.R, OpNames.RD)))
        if self.verbose >= 2:
            print('MPO time = ', time.perf_counter() - tx)

        # TE
        me = MovingEnvironment(mpo, mps, mps, "TE")
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        tx = time.perf_counter()
        me.init_environments(self.verbose >= 3)
        if self.verbose >= 2:
            print('TE INIT time = ', time.perf_counter() - tx)
        te = TimeEvolution(me, VectorUBond(bond_dims), method, n_sub_sweeps)
        te.iprint = self.verbose
        te.solve(n_steps, beta_step, mps.center == 0)

        self.bond_dim = bond_dims[-1]
        mps.save_data()
        mpo.deallocate()
        mps_info.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE imaginary time evolution | Time = %.2f <<<' %
                  (time.perf_counter() - t))

    def decompression(self, bond_dim):
        if self.verbose >= 2:
            print('>>> START decompression <<<')
        t = time.perf_counter()

        # Ancilla MPSInfo (thermal)
        mps_info_thermal = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                          self.target, self.hamil.basis)
        mps_info_thermal.tag = "INIT"

        # Ancilla MPSInfo (initial)
        mps_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum, self.target,
                                  self.hamil.basis)
        mps_info.set_bond_dimension(bond_dim)
        mps_info.tag = "INIT2"
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        # Ancilla MPS (thermal)
        mps_info_thermal.load_mutable()
        mps_thermal = MPS(mps_info_thermal)
        mps_thermal.load_data()
        mps_thermal.load_mutable()

        # Ancilla MPS (initial)
        mps = MPS(self.n_sites, self.n_sites - 2, 2)
        mps_info.load_mutable()
        mps.initialize(mps_info)
        mps.random_canonicalize()

        # MPS/MPSInfo save mutable
        mps.save_mutable()
        mps.deallocate()
        mps_info.deallocate_mutable()
        mps_thermal.save_mutable()
        mps_thermal.deallocate()
        mps_info_thermal.deallocate_mutable()

        # Identity MPO
        impo = IdentityMPO(self.hamil)
        impo = AncillaMPO(impo)
        impo = SimplifiedMPO(impo, Rule())

        # ME
        ime = MovingEnvironment(impo, mps, mps_thermal, "COMPRESS")
        ime.init_environments()

        # Linear
        cps = Linear(ime, VectorUBond(
            [bond_dim]), VectorUBond([10]), VectorDouble([0.0]))
        cps.solve(30, False)

        mps_info.load_mutable()
        mps.load_mutable()
        mps_info.tag = "INIT"
        mps.save_data()
        mps_info.save_mutable()
        mps.save_mutable()
        mps.deallocate()
        mps_info.deallocate_mutable()
        impo.deallocate()
        mps_info.deallocate()
        mps_info_thermal.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE decompression mps | Time = %.2f <<<' %
                  (time.perf_counter() - t))

    # particle number correlation
    # return value (SU2):
    #     npc[0, :, :] -> <(N_{i,alpha}+N_{i,beta}) (N_{j,alpha}+N_{j,beta})>
    # return value (SZ):
    #     npc[0, :, :] -> <N_{i,alpha} N_{j,alpha}>
    #     npc[1, :, :] -> <N_{i,alpha}  N_{j,beta}>
    #     npc[2, :, :] -> < N_{i,beta} N_{j,alpha}>
    #     npc[3, :, :] -> < N_{i,beta}  N_{j,beta}>
    def get_one_npc(self, ridx=None):
        if self.verbose >= 2:
            print('>>> START one-npc <<<')
        t = time.perf_counter()

        self.hamil.mu = 0.0

        # Ancilla MPSInfo (final)
        mps_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                  self.target, self.hamil.basis)
        mps_info.tag = "FINAL"

        # Ancilla MPS (final)
        mps = MPS(mps_info)
        mps.load_data()

        # 1NPC MPO
        pmpo = NPC1MPOQC(self.hamil)
        pmpo = AncillaMPO(pmpo, True)
        pmpo = SimplifiedMPO(pmpo, Rule())

        # 1NPC
        pme = MovingEnvironment(pmpo, mps, mps, "1NPC")
        pme.init_environments(self.verbose >= 3)
        expect = Expect(pme, self.bond_dim, self.bond_dim)
        expect.iprint = self.verbose
        expect.solve(True, mps.center == 0)
        if SpinLabel == SU2:
            dmr = expect.get_1npc_spatial(0, self.n_physical_sites)
            dm = np.array(dmr).copy()
        else:
            dmr = expect.get_1npc(0, self.n_physical_sites)
            dm = np.array(dmr).copy()
            dm = dm.reshape((self.n_physical_sites, 2,
                             self.n_physical_sites, 2))
            dm = np.transpose(dm, (0, 2, 1, 3))

        if ridx is not None:
            dm[:, :] = dm[ridx, :][:, ridx]

        mps.save_data()
        dmr.deallocate()
        pmpo.deallocate()
        mps_info.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE one-npc | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        if SpinLabel == SU2:
            return dm[None, :, :]
        else:
            return np.concatenate([dm[None, :, :, 0, 0], dm[None, :, :, 0, 1],
                                   dm[None, :, :, 1, 0], dm[None, :, :, 1, 1]], axis=0)

    # one-particle density matrix
    # return value:
    #     pdm[0, :, :] -> <AD_{i,alpha} A_{j,alpha}>
    #     pdm[1, :, :] -> < AD_{i,beta}  A_{j,beta}>
    def get_one_pdm(self, ridx=None):
        if self.verbose >= 2:
            print('>>> START one-pdm <<<')
        t = time.perf_counter()

        self.hamil.mu = 0.0

        # Ancilla MPSInfo (final)
        mps_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                  self.target, self.hamil.basis)
        mps_info.tag = "FINAL"

        # Ancilla MPS (final)
        mps = MPS(mps_info)
        mps.load_data()

        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil)
        pmpo = AncillaMPO(pmpo, True)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 1PDM
        pme = MovingEnvironment(pmpo, mps, mps, "1PDM")
        pme.init_environments(self.verbose >= 3)
        expect = Expect(pme, self.bond_dim, self.bond_dim)
        expect.iprint = self.verbose
        expect.solve(True, mps.center == 0)
        if SpinLabel == SU2:
            dmr = expect.get_1pdm_spatial(self.n_physical_sites)
            dm = np.array(dmr).copy()
        else:
            dmr = expect.get_1pdm(self.n_physical_sites)
            dm = np.array(dmr).copy()
            dm = dm.reshape((self.n_physical_sites, 2,
                             self.n_physical_sites, 2))
            dm = np.transpose(dm, (0, 2, 1, 3))

        if ridx is not None:
            dm[:, :] = dm[ridx, :][:, ridx]

        mps.save_data()
        dmr.deallocate()
        pmpo.deallocate()
        mps_info.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE one-pdm | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        if SpinLabel == SU2:
            return np.concatenate([dm[None, :, :], dm[None, :, :]], axis=0) / 2
        else:
            return np.concatenate([dm[None, :, :, 0, 0], dm[None, :, :, 1, 1]], axis=0)

    # two-particle density matrix (SZ only)
    # return value:
    #     pdm[0, i, j, k, l] -> <AD_{i,alpha} AD_{j,alpha} A_{k,alpha} A_{l,alpha}>
    #     pdm[1, i, j, k, l] -> <AD_{i,alpha} AD_{j, beta} A_{k, beta} A_{l,alpha}>
    #     pdm[2, i, j, k, l] -> <AD_{i, beta} AD_{j, beta} A_{k, beta} A_{l, beta}>
    def get_two_pdm(self, ridx=None):
        if self.verbose >= 2:
            print('>>> START two-pdm <<<')
        t = time.perf_counter()

        self.hamil.mu = 0.0

        # only support SZ
        assert SpinLabel == SZ

        # Ancilla MPSInfo (final)
        mps_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                  self.target, self.hamil.basis)
        mps_info.tag = "FINAL"

        # Ancilla MPS (final)
        mps = MPS(mps_info)
        mps.load_data()

        # 2PDM MPO
        pmpo = PDM2MPOQC(self.hamil, mask=PDM2MPOQC.s_minimal)
        pmpo = AncillaMPO(pmpo, True)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 2PDM
        pme = MovingEnvironment(pmpo, mps, mps, "2PDM")
        pme.init_environments(self.verbose >= 3)
        expect = Expect(pme, self.bond_dim, self.bond_dim)
        expect.iprint = self.verbose
        expect.solve(True, mps.center == 0)

        dmr = expect.get_2pdm(self.n_physical_sites)
        dm = np.array(dmr, copy=True)
        dm = dm.reshape((self.n_physical_sites, 2,
                         self.n_physical_sites, 2, self.n_physical_sites, 2, self.n_physical_sites, 2))
        dm = np.transpose(dm, (0, 2, 4, 6, 1, 3, 5, 7))

        if ridx is not None:
            dm[:, :, :, :] = dm[ridx, :, :, :][:, ridx,
                                               :, :][:, :, ridx, :][:, :, :, ridx]

        mps.save_data()
        pmpo.deallocate()
        mps_info.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE two-pdm | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        return np.concatenate([dm[None, :, :, :, :, 0, 0, 0, 0], dm[None, :, :, :, :, 0, 1, 1, 0],
                               dm[None, :, :, :, :, 1, 1, 1, 1]], axis=0)

    def __del__(self):
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        release_memory()


if __name__ == "__main__":

    # parameters
    bond_dim = 400
    beta = 1.0
    beta_step = 0.2
    mu = -1.0
    bond_dims = [bond_dim]
    n_threads = 8
    hf_type = "RHF"
    mpg = 'd2h'
    pg_reorder = True
    scratch = '/central/scratch/hczhai/hchain'
    scratch = './nodex'
    scratch = '/scratch/local/hczhai/hchain'

    import os
    if not os.path.isdir(scratch):
        os.mkdir(scratch)
    os.environ['TMPDIR'] = scratch

    from pyscf import gto, scf, symm, ao2mo

    # H chain
    N = 8
    BOHR = 0.52917721092  # Angstroms
    R = 1.8 * BOHR
    mol = gto.M(atom=[['H', (i * R, 0, 0)] for i in range(N)],
                basis='sto6g', verbose=0, symmetry=mpg)
    pg = mol.symmetry.lower()

    # Reorder
    if pg == 'd2h':
        fcidump_sym = ["Ag", "B3u", "B2u", "B1g", "B1u", "B2g", "B3g", "Au"]
        optimal_reorder = ["Ag", "B1u", "B3u",
                           "B2g", "B2u", "B3g", "B1g", "Au"]
    elif pg == 'c1':
        fcidump_sym = ["A"]
        optimal_reorder = ["A"]
    else:
        raise FTDMRGError("Point group %d not supported yet!" % pg)

    if hf_type == "RHF":
        # SCF
        m = scf.RHF(mol)
        m.kernel()
        mo_coeff = m.mo_coeff
        n_ao = mo_coeff.shape[0]
        n_mo = mo_coeff.shape[1]

        orb_sym_str = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff)
        orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])

        # Sort the orbitals by symmetry for more efficient DMRG
        if pg_reorder:
            idx = np.argsort([optimal_reorder.index(i) for i in orb_sym_str])
            orb_sym = orb_sym[idx]
            mo_coeff = mo_coeff[:, idx]
            ridx = np.argsort(idx)
        else:
            ridx = np.array(list(range(n_mo)), dtype=int)

        h1e = mo_coeff.T @ m.get_hcore() @ mo_coeff
        g2e = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff), n_mo)
        ecore = mol.energy_nuc()
        ecore = 0.0

    elif hf_type == "UHF":
        assert SpinLabel == SZ
        # SCF
        m = scf.UHF(mol)
        m.kernel()
        mo_coeff_a, mo_coeff_b = m.mo_coeff[0], m.mo_coeff[1]
        n_ao = mo_coeff_a.shape[0]
        n_mo = mo_coeff_b.shape[1]

        orb_sym_str_a = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff_a)
        orb_sym_str_b = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff_b)
        orb_sym_a = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_a])
        orb_sym_b = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_b])

        # Sort the orbitals by symmetry for more efficient DMRG
        if pg_reorder:
            idx_a = np.argsort([optimal_reorder.index(i)
                                for i in orb_sym_str_a])
            orb_sym_a = orb_sym_a[idx_a]
            mo_coeff_a = mo_coeff_a[:, idx_a]
            idx_b = np.argsort([optimal_reorder.index(i)
                                for i in orb_sym_str_b])
            orb_sym_b = orb_sym_b[idx_b]
            mo_coeff_b = mo_coeff_b[:, idx_b]
            assert np.allclose(idx_a, idx_b)
            assert np.allclose(orb_sym_a, orb_sym_b)
            orb_sym = orb_sym_a
            ridx = np.argsort(idx_a)
        else:
            orb_sym = orb_sym_a
            ridx = np.array(list(range(n_mo)), dtype=int)

        h1ea = mo_coeff_a.T @ m.get_hcore() @ mo_coeff_a
        h1eb = mo_coeff_b.T @ m.get_hcore() @ mo_coeff_b
        g2eaa = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_a), n_mo)
        g2ebb = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_b), n_mo)
        g2eab = ao2mo.kernel(
            mol, [mo_coeff_a, mo_coeff_a, mo_coeff_b, mo_coeff_b])
        h1e = (h1ea, h1eb)
        g2e = (g2eaa, g2ebb, g2eab)
        ecore = mol.energy_nuc()
        ecore = 0.0

    ft = FTDMRG(scratch=scratch, memory=10E9, verbose=2, omp_threads=n_threads)
    ft.init_hamiltonian(pg, n_sites=n_mo, twos=0, isym=1,
                        orb_sym=orb_sym, e_core=ecore, h1e=h1e, g2e=g2e)
    ft.generate_initial_mps(bond_dim)
    # when reorder == False, for some bad ordering
    # it is hard for sweep algorithm to recover missing quantum numbers
    # then 1pdm will have large error
    # decompression can alleviate this problem
    ft.decompression(bond_dim)
    n_steps = int(round(beta / beta_step) + 0.1)

    # sub-division first-step (log)
    # ng = 10
    # ngx = [1]
    # for i in range(1, ng):
    #     ngx.append(ngx[-1] * 2)
    # ngxs = sum(ngx)
    # stp = [float(x) / ngxs for x in ngx]
    # for i in range(0, ng):
    #     ft.imaginary_time_evolution(1, beta_step / 2 * stp[i], mu, bond_dims, TETypes.TangentSpace, n_sub_sweeps=1, cont=i != 0)

    # ft.decompression(50)

    print("n_steps = ", n_steps)
    ft.imaginary_time_evolution(
        1, beta_step / 2, mu, bond_dims, TETypes.RK4, n_sub_sweeps=10)
    # after the first beta step, use 2 sweeps (or 1 sweep) for each beta step
    if n_steps != 1:
        ft.imaginary_time_evolution(
            n_steps - 1, beta_step / 2, mu, bond_dims, TETypes.RK4, n_sub_sweeps=2, cont=True)
    pdm1 = ft.get_one_pdm(ridx)
    npc1 = ft.get_one_npc(ridx)
    pdm2 = ft.get_two_pdm(ridx)

    import pickle
    pickle.dump(pdm1, open('pdm1', "wb"))
    pickle.dump(pdm2, open('pdm2', "wb"))

    del ft
